#! /bin/bash
# $Id: dig_wrappers.bash,v 1.5 2004/09/23 17:07:14 mszick Exp $

:<<-'__dig_wrappers_Doc_'

    Copyright (c) Michael S. Zick, 2004
    License: Unrestricted reuse in any form, for any purpose.
    Warranty: None -{Its a script; the user is on their own.}-

Provides
    Functions which act as wrappers around the
    external utility program 'dig'.

NOTE ! - ! - !
    1) Each lookup has its own function only for demonstration.
       This also provides a workspace to anyone that needs
       to rewrite the wrappers.

    2) Quotes and section references are to the html, ARM
       (Administrator Reference Manual), shipped with Bind
       version 9.3.0rc3

Compatibility
    The utility program 'dig' (Domain Information Groper) is
    part of the 'bind' set of programs.

    That set of programs are what provides DNS
    (Domain Name Service) to the Internet.
    See: <http://www.isc.org>

    There are three revision series of the 'bind' programs:

    Bind revision series 9 (current)
        Provides a 'dig' utility that supports the
        '+short' option.

        These wrappers are written for the version
        of 'dig' that is part of revision series 9.

    Bind revision series 8 (still in use)
        Provides a 'dig' utility.  This version does
        not support the '+short' option.

        Users with this version of 'dig' get to either
        upgrade their version of 'dig' or rewrite these
        wrappers to parse the replies into '+short' form.

    Bind revision series 4 (dead)
        Does not provide a 'dig' utility.

        Users with this version really, really, should
        upgrade their bind/named installation.

    All revision series
        The utility program 'nslookup' is present in all
        revision series (but soon to disappear).

        These wrappers would need to be rewritten to
        use the 'nslookup' utility.


__dig_wrappers_Doc_

# Special IFS settings used for string parsing.
# Whitespace == :Space:Tab:Line Feed:Carriage Return:
WSP_IFS=$'\x20'$'\x09'$'\x0A'$'\x0D'

# No Whitespace == Line Feed:Carriage Return
NO_WSP=$'\x0A'$'\x0D'

# Field separator for dotted decimal IP addresses
ADR_IFS=${NO_WSP}'.'

# To convert an array into a dotted string...
DOT_IFS='.'${WSP_IFS}

#### Helper functions ####

# This function described in is_number.bash
# is_number <input>
is_number() {
    [ $# -eq 1 ]      || return 1   # is blank?
    [ x"$1" == 'x0' ] && return 0   # is zero?
    local -i tst
    let tst=$1 2>/dev/null          # is numeric?
    return $?
}

# This function described in is_address.bash
# is_address <input>
is_address() {
    [ $# -eq 1 ] || return 1    # Blank ==> false

    local -a _ia_input
    local IFS=${ADR_IFS}
    _ia_input=( $1 )

    if  [ ${#_ia_input[@]} -eq 4 ]  &&
        is_number ${_ia_input[0]}   &&
        is_number ${_ia_input[1]}   &&
        is_number ${_ia_input[2]}   &&
        is_number ${_ia_input[3]}   &&
        [ ${_ia_input[0]} -lt 256 ] &&
        [ ${_ia_input[1]} -lt 256 ] &&
        [ ${_ia_input[2]} -lt 256 ] &&
        [ ${_ia_input[3]} -lt 256 ]
    then
        return 0
    else
        return 1
    fi
}

# This function described in dot_array.bash
# dot_array <array_name>
dot_array() {
    [ $# -eq 1 ] || return 1

    local -a _da_input
    eval  _da_input=\(\ \$\{$1\[@\]\}\ \)

    local IFS=${DOT_IFS}
    local _da_output=${_da_input[@]}
    IFS=${WSP_IFS}

    echo $_da_output
    return 0
}

# This function described in: split_ip.bash
# split_ip <IP_address> <array_name_norm> [<array_name_rev>]
split_ip() {
    [ $# -eq 3 ] ||              # Either three
    [ $# -eq 2 ] || return 1     # or two arguments

    # Parse by assignment into an array
    local -a _si_input
    local IFS=${ADR_IFS}
    _si_input=( $1 )
    IFS=${WSP_IFS}

    # 'Copy by name' as the normal order array
    eval $2=\(\ \$\{_si_input\[@\]\}\ \)

    if [ $# -eq 3 ]
    then
        # Build query order array
        local -a _dns_ip
        _dns_ip[0]=${_si_input[3]}
        _dns_ip[1]=${_si_input[2]}
        _dns_ip[2]=${_si_input[1]}
        _dns_ip[3]=${_si_input[0]}
        eval $3=\(\ \$\{_dns_ip\[@\]\}\ \)
    fi
    return 0
}

#### Blacklist server, dig wrappers ####

# Get the dns text resource record(s)
# from a blacklist name server.

# The text record(s) from a blacklist server
# typically contain a hyperlink to the
# blacklist report(s).

# <dns_to_query> is the reversed octet IP address
# <list_to_query> is the blacklist server domain name

# get_list_text <dns_to_query> <list_to_query>
get_list_text() {
    [ $# -eq 2 ] || return 1

    # return code from external lookup
    local _glt_rc

    # array for multiple line replies
    local -a _glt_reply

    # Line-read mode
    IFS=${NO_WSP}

    # Forward lookup - a (very strange) name to IP text record
    # Short form answer, try all name servers, class=IN, type=TXT
    _glt_reply=( $(dig +short +nofail ${1}'.'${2} -c in -t txt) )

    # save program return code
    _glt_rc=$?

    # restore IFS mode
    IFS=${WSP_IFS}

    # progress report
    if [ ${_glt_rc} -ne 0 ]
    then
        echo -n 'External look-up failure code: '${_glt_rc}
        echo ' for text records at: '${1}'.'${2}
        return ${_glt_rc}
    else
        echo -n 'Blacklist server: '${2}' has '
        echo ${#_glt_reply[@]}' text records on file for: '${1}
        for (( _i = 0 ; _i < ${#_glt_reply[@]} ; _i++ ))
        do
            echo '    '${_glt_reply[$_i]}
        done
    fi
    return 0
}

####
#
# The other wrappers are similar
# Except for the get_ns_chain,
# which does its own reply parsing.
#
####

# Get the listing code from a blacklist server.

# None found == Not Listed

# Else the IP address returned is the
# listing code.

# <dns_to_query> is the reversed octet IP address
# <list_to_query> is the blacklist server domain name

# For testing only - a place for the last found error code
declare err_code

# get_list_code <dns_to_query> <list_to_query>
get_list_code() {
    [ $# -eq 2 ] || return 1

    local _glc_rc
    local -a _glc_reply

    IFS=${NO_WSP}
    # Forward lookup - a (very strange) name to IP address record
    # Short form answer, try all name servers, class=IN, type=A
    _glc_reply=( $(dig +short +nofail ${1}'.'${2} ) )
    _glc_rc=$?
    IFS=${WSP_IFS}

    if [ ${_glc_rc} -ne 0 ]
    then
        echo -n 'External look-up failure code: '${_glc_rc}
        echo ' for address records at: '${1}'.'${2}
        return ${_glc_rc}
    else
        echo -n 'Blacklist server: '${2}' has '
        echo ${#_glc_reply[@]}' listing codes on file for: '${1}
        for (( _i = 0 ; _i < ${#_glc_reply[@]} ; _i++ ))
        do
            if is_address ${_glc_reply[$_i]}
            then
                err_code=${_glc_reply[$_i]}
            fi
            echo '    '${_glc_reply[$_i]}
        done
    fi
    return 0

}

#### Basic name server queries ####

:<<-'__DNS_ZONE__'
6.3.1.1. Resource Records
    A domain name identifies a node.

    Each node has a set of resource information,
    which may be empty.

    The set of resource information associated with
    a particular name is composed of separate RRs.

6.3.1.2. Textual expression of RRs
    The resource data or RDATA section of the RR are given
    using knowledge of the typical representation for the data.

    (That quote refers to the ARM content, but dig
     translates to the same format.)

6.3.2. Discussion of MX Records
    MX records are used to control delivery of email.

    The data specified in the record is a priority
    and a domain name.

    The domain name given is the machine to which the
    mail will be delivered.
    It must have an associated A record — CNAME is not sufficient.

    For a given domain, if there is both a CNAME record and an
    MX record, the MX record is in error, and will be ignored.

6.3.4. Inverse Mapping in IPv4
    Reverse name resolution
    (that is, translation from IP address to name) is achieved
    by means of the in-addr.arpa domain and PTR records.

    ... optionally, multiple PTR records if the machine has
    more than one name.

    (The '-x' option of dig will take care of query formatting.)


__DNS_ZONE__

# Forward dns lookup - Domain name to IP address

# domain_name is a string holding the dotted name

# For testing only - a place for name and last found number
declare ip_name
declare ip_addr

# get_address_rec <domain_name>
get_address_rec() {
    [ $# -eq 1 ] || return 1

    local _gar_rc
    local -a _gar_reply

    IFS=${NO_WSP}
    # Forward lookup - a domain name to IP address record
    # Short form answer, try all name servers, class=IN, type=A
    _gar_reply=( $(dig +short +nofail ${1} -c in -t a) )
    _gar_rc=$?
    IFS=${WSP_IFS}

    if [ ${_gar_rc} -ne 0 ]
    then
        echo -n 'External look-up failure code: '${_gar_rc}
        echo    ' for address records of: '${1}
        return ${_gar_rc}
    else
        echo -n 'Found '${#_gar_reply[@]}
        echo    ' address records on file for: '${1}
        for (( _i = 0 ; _i < ${#_gar_reply[@]} ; _i++ ))
        do
            if is_address ${_gar_reply[$_i]}
            then
                ip_addr=${_gar_reply[$_i]}
            fi
            echo '    '${_gar_reply[$_i]}
        done
    fi
    return 0
}

# Reverse dns lookup - IP address to Domain name(s)

# ip_address is a normal order set of octets;
# dig will reformat them (-x) for us.

# get_name_rec <ip_address>
get_name_rec() {
    [ $# -eq 1 ] || return 1

    local _gnr_rc
    local -a _gnr_reply

    IFS=${NO_WSP}
    # Reverse lookup - IP address to name record(s)
    # Short form answer, try all name servers, class=IN, type=A
    _gnr_reply=( $(dig +short +nofail -x ${1} -c in -t ptr) )
    _gnr_rc=$?
    IFS=${WSP_IFS}

    if [ ${_gnr_rc} -ne 0 ]
    then
        echo -n 'External look-up failure code: '${_gnr_rc}
        echo    ' for name records of: '${1}
        return ${_gnr_rc}
    else
        echo -n 'Found '${#_gnr_reply[@]}
        echo    ' name records on file for: '${1}
        for (( _i = 0 ; _i < ${#_gnr_reply[@]} ; _i++ ))
        do
            echo '    '${_gnr_reply[$_i]}
        done
    fi
    return 0
}

# Domain name SOA (Start Of Authority) lookup
# Every named node on the Internet has one of these,
# (or should) - it is what defines an Internet node.

# get_start_authority <domain_name>
get_start_authority() {
    [ $# -eq 1 ] || return 1

    local _gsa_rc
    local -a _gsa_reply # Should ony be one line

    IFS=${NO_WSP}
    # Reverse lookup - IP address to name record(s)
    # Short form answer, try all name servers, class=IN, type=A
    _gsa_reply=( $(dig +noall +nofail +authority ${1} -c in -t soa) )
    _gsa_rc=$?
    IFS=${WSP_IFS}

    if [ ${_gsa_rc} -ne 0 ]
    then
        echo -n 'External look-up failure code: '${_gsa_rc}
        echo    ' for start of authority records of: '${1}
        return ${_gsa_rc}
    else
        echo -n 'Found '${#_gsa_reply[@]}
        echo    ' start of authority records on file for: '${1}
        for (( _i = 0 ; _i < ${#_gsa_reply[@]} ; _i++ ))
        do
            echo '    '${_gsa_reply[$_i]}
        done
    fi
    return 0
}

#### Spammer network look-ups ####

# The point of 'spam' is to get a reply from
# someone.
# So we want a list of the mail transports.

# These are all forward lookups which return
# a domain name (not an IP address)

# domain_name may be a fully qualified name.
# get_mx_rec <domain_name>
get_mx_rec() {
    [ $# -eq 1 ] || return 1

    local _gmxr_rc
    local -a _gmxr_reply

    IFS=${NO_WSP}
    # Forward lookup; Domain_Name to Mail_Exchange_Name
    # Short form answer, try all name servers, class=IN, type=MX
    _gmxr_reply=( $(dig +short +nofail ${1} -c in -t mx) )
    _gmxr_rc=$?
    IFS=${WSP_IFS}

    if [ ${_gmxr_rc} -ne 0 ]
    then
        echo -n 'External look-up failure code: '${_gmxr_rc}
        echo    ' for mail exchange records of: '${1}
        return ${_gmxr_rc}
    else
        echo -n 'Found '${#_gmxr_reply[@]}
        echo    ' mail exchange records on file for: '${1}
        for (( _i = 0 ; _i < ${#_gmxr_reply[@]} ; _i++ ))
        do
            echo '    '${_gmxr_reply[$_i]}
        done
    fi
    return 0
}

# Setting up a list of mail transports, or
# other URL's means there is a name server out there.
# So we want a list of them also.

# For testing only - the first nameserver name
declare ns_name

# domain_name may be a fully qualified name.
# get_ns_rec <domain_name>
get_ns_rec() {
    [ $# -eq 1 ] || return 1

    local _gnsr_rc
    local -a _gnsr_reply

    IFS=${NO_WSP}
    # Forward lookup - Domain_Name to Name_Server_Name.
    # Short form answer, try all name servers, class=IN, type=NS
    _gnsr_reply=( $(dig +short +nofail ${1} -c in -t ns) )
    _gnsr_rc=$?
    IFS=${WSP_IFS}

    if [ ${_gnsr_rc} -ne 0 ]
    then
        echo -n 'External look-up failure code: '${_gnsr_rc}
        echo    ' for name server records of: '${1}
        return ${_gnsr_rc}
    else
        ns_name=${_gnsr_reply[0]}
        echo -n 'Found '${#_gnsr_reply[@]}
        echo    ' name server records on file for: '${1}

        if [ ${#_gnsr_reply[@]} -eq 0 ]
        then
            return 1
        fi
        for (( _i = 0 ; _i < ${#_gnsr_reply[@]} ; _i++ ))
        do
            echo '    '${_gnsr_reply[$_i]}
        done
    fi
    return 0
}

# If a standard check for a name server record fails,
# then trace the domain name authority list...
# Expect this to be (perhaps intentionally) a broken
# chain of authority...
# So just wait for 'dig' to timeout somewhere.

# Note: This does not use the '+short' option,
# so the reply will require a lot of parsing.

# get_ns_chain <domain_name>
get_ns_chain(){
    [ $# -eq 1 ] || return 1

    local _gnsc_rc
    local -a _gnsc_reply
    local -a _gnsc_line

    local -i _gnsc_top
    local -i _gnsc_tmp

    local -i _gnsc_found=0

    IFS=${NO_WSP}
    # Name server authority chain trace
    # Method: Brute force lookup everything.
    echo ' + + + Please wait - This needs to time-out. + + +'
    _gnsc_reply=( $(dig +nofail +all +trace ${1}) )
    _gnsc_rc=$?

    IFS=${WSP_IFS}

    if [ ${#_gnsc_reply[@]} -eq 0 ]
    then
        echo 'No reply from authority chain request.'
        echo 'Query authority chain of: '${1}
        return 1
    fi

    if [ ${_gnsc_rc} -ne 0 ]
    then
        echo 'Authority chain broken. Return code: '${_gnsc_rc}
    fi

    _gnsc_top=${#_gnsc_reply[@]}-1
    # This works because we know there is more than
    # a single line in the reply at this point.
    # I.E: The Internet root nameservers always answer.
    while [ ${_gnsc_top} -ne 0 ]
    do
        _gnsc_line=( ${_gnsc_reply[${_gnsc_top}]} )
        _gnsc_tmp=${#_gnsc_line[@]}
        if [ x"${_gnsc_line[0]}" != 'x;;' ]
        then
            if [ ${_gnsc_tmp} -gt 3 ] &&
               [ x"${_gnsc_line[3]}" == 'xNS' ]
            then
                echo '    '${_gnsc_line[4]}' is serving <'${_gnsc_line[0]}'>'
                _gnsc_found=1
                ns_name=${_gnsc_line[4]}
            fi
        else
            if [ ${_gnsc_found} -ne 0 ]
            then
                return 0
            fi
        fi
        _gnsc_top=${_gnsc_top}-1
    done
}

# It is very likely that the nameserver is running 'bind'.
# Query for nameserver information (which may be hidden).

# domain_nameserver is the full name of the nameserver.
# get_ns_rec <domain_nameserver>
get_ns_text() {
    [ $# -eq 1 ] || return 1

    local _rc_version
    local _rc_hostname
    local _rc_server_id
    local _gnst_rc
    local -a _gnst_version
    local -a _gnst_hostname
    local -a _gnst_server_id

    IFS=${NO_WSP}
    # Information lookup of the nameserver program
    # Specify the authoritative server name,
    # Short form answer, servername, query, class=CHOAS, type=TXT
    _gnst_version=( $(dig +short @${1} version.bind -c ch -t txt) )
    _rc_version=$?
    _gnst_hostname=( $(dig +short @${1} hostname.bind -c ch -t txt) )
    _rc_hostname=$?
    _gnst_server_id=( $(dig +short @${1} id.server -c ch -t txt) )
    _rc_server_id=$?
    IFS=${WSP_IFS}

    _gnst_rc=0
    if [ ${_rc_version} -ne 0 ]
    then
        echo -n 'External look-up failure code: '${_rc_version}
        echo    ' for name server version records at: '${1}
        _gnst_rc=$_rc_version
    else
        echo -n 'Found '${#_gnst_version[@]}
        echo    ' name server version records on file for: '${1}
        for (( _i = 0 ; _i < ${#_gnst_version[@]} ; _i++ ))
        do
            echo '    '${_gnst_version[$_i]}
        done
    fi

    if [ ${_rc_hostname} -ne 0 ]
    then
        echo -n 'External look-up failure code: '${_rc_hostname}
        echo    ' for name server hostname records at: '${1}
        _gnst_rc=$_rc_hostname
    else
        echo -n 'Found '${#_gnst_hostname[@]}
        echo    ' name server hostname records on file for: '${1}
        for (( _i = 0 ; _i < ${#_gnst_hostname[@]} ; _i++ ))
        do
            echo '    '${_gnst_hostname[$_i]}
        done
    fi

    if [ ${_rc_server_id} -ne 0 ]
    then
        echo -n 'External look-up failure code: '${_rc_server_id}
        echo    ' for name server server_id records at: '${1}
        _gnst_rc=$_rc_server_id
    else
        echo -n 'Found '${#_gnst_server_id[@]}
        echo    ' name server server_id records on file for: '${1}
        for (( _i = 0 ; _i < ${#_gnst_server_id[@]} ; _i++ ))
        do
            echo '    '${_gnst_server_id[$_i]}
        done
    fi
    return $_gnst_rc
}

#### Test code

# Given a domain name,
# try all functions on it.

# Output can be checked against:
# dig domain.name -c in -t any

# place for the forward order and reverse order octets
declare -a fwd_dns_components
declare -a rev_dns_components

# place for the forward order and reverse order
# dotted IP address
declare fwd_str
declare rev_str

# single (free) blacklist server to use
declare bl_server='sbl-xbl.spamhaus.org'

if [ $# -ne 1 ]
then
    echo 'This example requires a single argument.'
    echo 'The argument should be a domain name.'
    exit 1
fi

ip_name=$1

echo
get_start_authority $ip_name

echo
if get_address_rec $ip_name
then
    echo
    if is_address $ip_addr
    then
        split_ip $ip_addr fwd_dns_components rev_dns_components
        fwd_str=$(dot_array fwd_dns_components)
        rev_str=$(dot_array rev_dns_components)

        get_name_rec $ip_addr
    fi

    echo
    if get_ns_rec $ip_name
    then
        echo
        get_ns_text $ns_name
    else
        echo
        get_ns_chain $ip_name
    fi

    echo
    get_mx_rec $ip_name

    echo
    if is_address $rev_str
    then
        get_list_code $rev_str $bl_server
        if is_address $err_code
        then
            echo
            get_list_text $rev_str $bl_server
        fi
    fi
fi

echo

:<<-'_dig_wrapper_outputs_'

> ./dig_wrappers.bash chunky.actsfast.info

Found 1 address records on file for: chunky.actsfast.info
    219.148.49.245

Found 0 name records on file for: 219.148.49.245

Found 0 name server records on file for: chunky.actsfast.info

 + + + Please wait - This needs to time-out. + + +
Authority chain broken. Return code: 9
    ns1.actsfast.info. is serving <actsfast.info.>
    ns2.actsfast.info. is serving <actsfast.info.>

Found 0 mail exchange records on file for: chunky.actsfast.info

Blacklist server: sbl-xbl.spamhaus.org has 1 listing codes on file for: 245.49.148.219
    127.0.0.2

Blacklist server: sbl-xbl.spamhaus.org has 1 text records on file for: 245.49.148.219
    "http://www.spamhaus.org/SBL/sbl.lasso?query=SBL17957"

_dig_wrapper_outputs_

exit 0