#! /bin/bash
# $Id: dot_array.bash,v 1.5 2004/09/14 22:36:30 mszick Exp $

:<<-'__dot_array_Doc_'

    Copyright (c) Michael S. Zick, 2004
    License: Unrestricted reuse in any form, for any purpose.
    Warranty: None -{Its a script; the user is on their own.}-

Provides
    Given the name of an array;
    Return on standard output the array elements
    as a string using '.' as field separators.

__dot_array_Doc_

# Special IFS settings used for string parsing.
# Whitespace == :Space:Tab:Line Feed:Carriage Return:
WSP_IFS=$'\x20'$'\x09'$'\x0A'$'\x0D'

# No Whitespace == Line Feed:Carriage Return
NO_WSP=$'\x0A'$'\x0D'

# Field separator for dotted decimal IP addresses
ADR_IFS=${NO_WSP}'.'

####

# The behavior of IFS (Input Field Separator) is
# to replace any single character that appears in
# the IFS string with the FIRST character of the
# IFS string.

# For any sequence of two or more IFS characters;
# substitute a single IFS[0] character.

# So to convert an array into a dotted string...
DOT_IFS='.'${WSP_IFS}
# Should do what we want.

####

# Output the array named in argument 1
# as a dotted string on stdout.
# dot_array <array_name>
dot_array() {
    [ $# -eq 1 ] || return 1     # Single argument required

    # Copy by name into local array
    local -a _da_input
    eval _da_input=\(\ \$\{$1\[@\]\}\ \)

    # Set the IFS to use '.' as a field separator
    local IFS=${DOT_IFS}

    # Parse the array by assignment into local string
    local _x_output=${_da_input[@]}
    IFS=${WSP_IFS}

    # Now output that on stdout so it can be used
    # in assignment statements.
    echo ${_da_output}

    return 0
}

####

# This function described in: split_ip.bash
# split_ip <IP_address> <array_name_norm> [<array_name_rev>]
split_ip() {
    [ $# -eq 3 ] ||              # Either three
    [ $# -eq 2 ] || return 1     # or two arguments

    # Parse by assignment into an array
    local -a _in_put
    local IFS=${ADR_IFS}
    _in_put=( $1 )
    IFS=${WSP_IFS}

    # 'Copy by name' as the normal order array
    eval $2=\(\ \$\{_in_put\[@\]\}\ \)

    if [ $# -eq 3 ]
    then
        # Build query order array
        local -a _dns_ip
        _dns_ip[0]=${_in_put[3]}
        _dns_ip[1]=${_in_put[2]}
        _dns_ip[2]=${_in_put[1]}
        _dns_ip[3]=${_in_put[0]}
        eval $3=\(\ \$\{_dns_ip\[@\]\}\ \)
    fi
    return 0
}

#### Test code

# A global array to hold IP address octets
declare -a ip_octets

# Fill that array using our helper function
split_ip 128.64.32.16 ip_octets

# Is the contents as expected?
echo
echo -n 'Test array of '${#ip_octets[@]}
echo    ' holding <'${ip_octets[@]}'>'

# See what our new function prints
echo
echo -n 'The dot_array function output is: '
dot_array ip_octets

# Which means we can use it in an assignment
# by specifying the result of the function.
dotted_octets=$(dot_array ip_octets)
echo
echo -n 'The saved result of dot_array is: '
echo $dotted_octets
echo

:<<-'__dot_array_outputs_'

Test array of 4 holding <128 64 32 16>

The dot_array function output is: 128.64.32.16

The saved result of dot_array is: 128.64.32.16

__dot_array_outputs_

exit 0