#! /bin/bash
# $Id: edit_by_glob.bash,v 1.4 2004/09/17 17:31:53 mszick Exp $

:<<-'__edit_by_glob_Doc_'

    Copyright (c) Michael S. Zick, 2004
    License: Unrestricted reuse in any form, for any purpose.
    Warranty: None -{Its a script; the user is on their own.}-

Provides
    A means to edit one array by the glob patterns within another.
    Effectively subtracts one array from the other by glob.

    Edit is based on a left anchored, glob match;
    Edit presumes that multiple matches are possible.

    If called with three arguments, exclude items are suffixed
    with an anything ('*') glob.
    This gives an 'edit by prefix match' without building
    glob patterns to exclude.

__edit_by_glob_Doc_

# Special IFS settings used for string parsing.
# Whitespace == :Space:Tab:Line Feed:Carriage Return:
WSP_IFS=$'\x20'$'\x09'$'\x0A'$'\x0D'

# No Whitespace == Line Feed:Carriage Return
NO_WSP=$'\x0A'$'\x0D'

# Edit target array by glob matching with exclude globs.
# Multiple matches in target are possible, no early exit.

# Providing fixed string exclude items (no globs) and
# calling with three arguments;
# Causes each exclude item to be suffixed with '*' giving
# an 'edit by prefix match' function.

# excludes_array_name - the array holding the element
# globs to remove. Globs are 'left anchored'.

# target_array_name - the array to edit

# edit_by_glob <excludes_array_name> <target_array_name>
edit_by_glob() {
    [ $# -eq 2 ] ||
    [ $# -eq 3 ] || return 1

    # Local copy of the exclude globs
    local -a _ebg_Excludes

    # Local copy of the target elements
    local -a _ebg_Target

    local _ebg_x
    local _ebg_t

    # Line-read mode
    local IFS=${NO_WSP}

    # No glob expansion
    set -f

    # Subscript pack exclude array by assignment
    eval _ebg_Excludes=\( \$\{$1\[@\]\} \)

    # Subscript pack exclude array by assignment
    eval _ebg_Target=\( \$\{$2\[@\]\} \)

    local _ebg_len=${#_ebg_Target[@]}      # Original list length
    local _ebg_cnt=${#_ebg_Excludes[@]}    # Exclude list length

    for (( x = 0; x < ${_ebg_cnt} ; x++ ))
    do
        _ebg_x=${_ebg_Excludes[$x]}
        [ $# -eq 3 ] && _ebg_x=${_ebg_x}'*'
        for (( n = 0 ; n < ${_ebg_len} ; n++ ))
        do
            if [ ${_ebg_Target[$n]:=} ]    # Defined & not empty
            then
                _ebg_t=${_ebg_Target[$n]/#${_ebg_x}/}
                [ ${#_ebg_t} -eq 0 ] && unset _ebg_Target[$n]
            fi
        done
    done

    # Subscript pack by assignment the result
    eval $2=\( \$\{_ebg_Target\[@\]\} \)
    set +f

    return 0
}


#### Test code ####
declare -a excludes
declare -a target

excludes=( '*.com' '*.net' )
target=( 'quickbucks.con' 'yahoo.com' 'sneaky.biz' 'goggle.com')

edit_by_glob excludes target

# What did that do for us?
echo
echo 'Test one, edit by glob.'
echo

for (( i = 0 ; i < ${#target[@]} ; i++ ))
do
    echo 'The target list element: '$i' contains: >'${target[$i]}'<'
done
echo

excludes=( 'ftp.' 'www.' )
target=( 'quickbucks.con' 'ftp.yahoo.com' 'sneaky.biz' 'www.goggle.com')

edit_by_glob excludes target prefix

# What did that do for us?
echo
echo 'Test two, edit by prefix.'
echo

for (( i = 0 ; i < ${#target[@]} ; i++ ))
do
    echo 'The target list element: '$i' contains: >'${target[$i]}'<'
done
echo

:<<-'_edit_by_glob_outputs_'

Test one, edit by glob.

The target list element: 0 contains: >quickbucks.con<
The target list element: 1 contains: >sneaky.biz<


Test two, edit by prefix.

The target list element: 0 contains: >quickbucks.con<
The target list element: 1 contains: >sneaky.biz<

_edit_by_glob_outputs_

exit 0