#! /bin/bash
# $Id: edit_exact.bash,v 1.4 2004/09/17 14:27:32 mszick Exp $

:<<-'__edit_exact_Doc_'

    Copyright (c) Michael S. Zick, 2004
    License: Unrestricted reuse in any form, for any purpose.
    Warranty: None -{Its a script; the user is on their own.}-

Provides
    A means to edit one array by the contents of another.
    Effectively subtracts one array from the other by contents.

    Edit is based on an exact, string, match;
    Edit presumes that only the first occurrence needs to be
    removed (unique target list entries).

    If called with a third (anything) argument, removes all
    occurrences of the exclude item from the target.

__edit_exact_Doc_

# Special IFS settings used for string parsing.
# Whitespace == :Space:Tab:Line Feed:Carriage Return:
WSP_IFS=$'\x20'$'\x09'$'\x0A'$'\x0D'

# No Whitespace == Line Feed:Carriage Return
NO_WSP=$'\x0A'$'\x0D'

# Edit target array by exact match with exclude array (no globbing)

# Removes first occurrence of edit list item, allowing early exit.
# ANYTHING as a third argument suppresses the early exit

# excludes_array_name - the array holding the elements to remove.
# target_array_name - the array to edit

# edit_exact <excludes_array_name> <target_array_name>
edit_exact() {
    [ $# -eq 2 ] ||
    [ $# -eq 3 ] || return 1

    # Local copy of the exclude array
    local -a _ee_Excludes

    # Local copy of the target array
    local -a _ee_Target

    local _ee_x
    local _ee_t

    # Line-read mode
    local IFS=${NO_WSP}

    # No glob expansion
    set -f

    # Subscript pack by assignment
    eval _ee_Excludes=\( \$\{$1\[@\]\} \)

    # Subscript pack by assignment
    eval _ee_Target=\( \$\{$2\[@\]\} \)

    local _ee_len=${#_ee_Target[@]}     # original length
    local _ee_cnt=${#_ee_Excludes[@]}   # Exclude list length

    # Can't edit zero length lists
    [ ${_ee_len} -ne 0 ] || return 0
    [ ${_ee_cnt} -ne 0 ] || return 0

    for (( x = 0; x < ${_ee_cnt} ; x++ ))
    do
        _ee_x=${_ee_Excludes[$x]}
        for (( n = 0 ; n < ${_ee_len} ; n++ ))
        do
            _ee_t=${_ee_Target[$n]}
            if [ x"${_ee_t}" == x"${_ee_x}" ]
            then
                unset _ee_Target[$n]    # Discard the match
                [ $# -eq 2 ] && break   # 2 arguments, done
            fi
        done
    done

    # Subscript sparse array packed by assignment.
      eval $2=\( \$\{_ee_Target\[@\]\} \)
    set +f

    return 0
}

#### Test code ####
declare -a exclude_list
declare -a target_list

exclude_list=( 'yahoo.com' 'google.com' )
base_list=( 'quickbucks.con' 'yahoo.com' 'sneaky.biz' 'google.com')

# Target list with duplicates
target_list=( ${base_list[@]} ${base_list[@]} )

echo
echo 'Test one - Remove first (or only) occurrence'
echo
edit_exact exclude_list target_list

# What did that do for us?

for (( i = 0 ; i < ${#target_list[@]} ; i++ ))
do
    echo 'The target list element: '$i' contains: >'${target_list[$i]}'<'
done

# Target list with duplicates
target_list=( ${base_list[@]} ${base_list[@]} )

echo
echo 'Test two - Remove all occurrences'
echo
edit_exact exclude_list target_list all

# What did that do for us?

for (( i = 0 ; i < ${#target_list[@]} ; i++ ))
do
    echo 'The target list element: '$i' contains: >'${target_list[$i]}'<'
done
echo

:<<-'_edit_exact_outputs_'

Test one - Remove first (or only) occurrence

The target list element: 0 contains: >quickbucks.con<
The target list element: 1 contains: >sneaky.biz<
The target list element: 2 contains: >quickbucks.con<
The target list element: 3 contains: >yahoo.com<
The target list element: 4 contains: >sneaky.biz<
The target list element: 5 contains: >google.com<

Test two - Remove all occurrences

The target list element: 0 contains: >quickbucks.con<
The target list element: 1 contains: >sneaky.biz<
The target list element: 2 contains: >quickbucks.con<
The target list element: 3 contains: >sneaky.biz<

_edit_exact_outputs_

exit 0