#! /bin/bash
# $Id: func_stack.bash,v 1.5 2004/09/23 17:12:01 mszick Exp $

:<<-'__func_stack_Doc_'

    Copyright (c) Michael S. Zick, 2003, 2004
    License: Unrestricted reuse in any form, for any purpose.
    Warranty: None -{Its a script; the user is on their own.}-

Provides
    A last-in-first-out stack for pending function calls.
    This stack machine may be used as the base on which
    to implement declarative program structures.

    Functions may have arguments.  Although the word
    'Function' is used here; actually any function or
    command, either Bash internal or external, may
    be placed on the pending operations stack.

    Any called function may pend additional function calls.

    Recursive algorithms may be emulated by a function
    pending itself for execution (rather than calling
    itself).

    Provides for grouped operations using a mark/release
    model of stack execution.

    This is the logic equivalent of using parentheses to
    group arithmetic expressions;
    where calling 'pend_mark' is the 'opening parenthesis'
    and calling 'pend_release' is the 'closing parenthesis'.

    Functions in a group may be an unconditional sequence
    of operations.

    Functions in a group may be a conditional sequence of
    operations by using the 'pend_drop' function call.

    Functions in a group would be a logical 'and' sequence;
    if any function which 'failed' calls 'pend_drop';
    dropping all pending operations remaining in the group.

    Functions in a group would be a logical 'or' sequence;
    if any function which 'succeeds' calls 'pend_drop';
    dropping all pending operations remaining in the group.

    The number of pending operations is limited by the
    maximum array subscript in Bash (about 32,000).

    Writing a Haskell or Prolog interpreter in Bash is
    left as a (very advanced) student exercise.

    Hints:
        You need a second stack like this one for
        backtrack functions (undo functions).

        When a function on this stack executes, it
        needs to pend its 'undo' function on the
        backtrack stack.

        Other notes on how to inter-lock these
        two stacks are in the comments.

    Happy declarative programming! (msz)

Bells_&_Whistles
    This version exposes all functions at the global level.
    A library version makes these methods of a '_pend' function.

    pend_mark
       Adds a 'pend_stop_mark' to the stack.
       This is the logic equivalent of an 'opening parenthesis'.

    pend_release
       Begins execution of pending operations at the top
       of the stack.
       This is the logic equivalent of a 'closing parenthesis'.

       Execution continues last-in-first-out until
       a 'pend_stop_mark' function is encountered.

       NOTE: Even a 'logically empty' stack has the
       'pend_stop_mark' function as its first entry.

    pend_init
       Empties the function stack of all entries;
       then pends the opening 'pend_stop_mark'.

    pend_func
       Adds the specified function_name or command_name
       to the stack, optionally with arguments.

       NOTE: Arguments must be escaped to survive the
       experience.  Use: $(printf '%q\n' 'argument' ...)

    pend_drop
       Drops stack entries down to the next
       'pend_stop_mark' function on the stack.

       NOTE: Even a 'logically empty' stack has the
       'pend_stop_mark' function as its first entry.

    Special purpose functions
       pend_stop_mark
           Used internally by pend_mark.
           May be called to escape infinite loops due
           to program logic errors.

       pend_dummy
           May be used to replace any function on the
           stack with a no-operation.
           Thou shall not let the stack become sparse
           sub-scripted by using 'unset ...' on it.

       pend_pop
           Drops the next function that would be executed.

       others
           The usual set of 'stack machine' operations
           can be added, but should not be required.

           The person writing a declarative language
           will want a save/restore of the stack
           to an external file.

Global Variables
    If your program directly accesses these - rethink
    your program - something is wrong with its logic.

    _pending_
        An array holding the actual stack of functions.
        Do not 'unset' any entries - assign 'pend_dummy'
        instead.

    _p_ctrl_
        An integer variable used to stop the stack runners.
        Setting to '0' is the same as calling
        'pend_stop_mark'; which should not be required.

    _pend_current_
        The function currently executing.
        Typically used in reports of what failed and
        trace logs to track what is executed.
        ("pend_func $_pend_current_" is recursion.)

__func_stack_Doc_

# Special IFS settings used for string parsing.
# Whitespace == :Space:Tab:Line Feed:Carriage Return:
WSP_IFS=$'\x20'$'\x09'$'\x0A'$'\x0D'

# No Whitespace == Line Feed:Carriage Return
NO_WSP=$'\x0A'$'\x0D'

# # # Pending operations # # #

# Global stack of pending operations.
declare -f -a _pending_

# Global sentinel for stack runners
declare -i _p_ctrl_

# Global holder for currently executing function
declare -f _pend_current_

# The do nothing function
pend_dummy() { : ; }

# Clear and initialize the function stack
pend_init() {
    unset _pending_[@]
    pend_func pend_stop_mark
    # pend_bt_func pend_bt_stop_mark
}

# Discard the top function on the stack
pend_pop() {
    if [ ${#_pending_[@]} -gt 0 ]
    then
        local -i _top_
        _top_=${#_pending_[@]}-1
        unset _pending_[$_top_]
    fi
}

# Push a function onto the stack
# pend_func function_name [$(printf '%q\n' arguments)]
pend_func() {
    local IFS=${NO_WSP}
    set -f
    _pending_[${#_pending_[@]}]=$@
    set +f

    # When the function just stacked
    # gets called; it should
    # 'pend_bt_func ...' I.E: its undo
    # function in case something later
    # in the group fails.
}

# The function which stops the release
pend_stop_mark() {
    _p_ctrl_=0
    # Group succeeded
    # pend_bt_drop
}

# Mark the stack - both 'pend_release'
# and 'pend_drop' stop running stack here
pend_mark() {
    pend_func pend_stop_mark
    # pend_bt_func pend_bt_stop_mark
}

# The two stack runners.
# Changed from the 2003 version to allow passing
# arguments and to call functions with a clean stack.

# Execute functions until 'pend_stop_mark'
pend_release() {
    local -i _top_

    # Protect from empty stack
    _p_ctrl_=${#_pending_[@]}

    while [ ${_p_ctrl_} -gt 0 ]
    do
       # Top is one less than count
       _top_=${#_pending_[@]}-1

       # Hold top function while adjusting stack
       _pend_current_=${_pending_[$_top_]}

       # Drop the top function
       unset _pending_[$_top_]

       # Now execute the saved top function
       eval $_pend_current_
    done
}

# Drop functions until 'pend_stop_mark'
pend_drop() {
    local -i _top_

    # Protect from empty stack
    local _pd_ctrl_=${#_pending_[@]}

    while [ ${_pd_ctrl_} -gt 0 ]
    do
       # Top is one less than count
       _top_=$_pd_ctrl_-1

       if [ "${_pending_[$_top_]}" == 'pend_stop_mark' ]
       then
           unset _pending_[$_top_]
           break
       else
           unset _pending_[$_top_]
           _pd_ctrl_=$_top_
       fi
    done

    # Protect from a physically empty stack
    if [ ${#_pending_[@]} -eq 0 ]
    then
        pend_func pend_stop_mark
    fi

    # Current group failed - undo things
    # pend_bt_release
}

#### Functions used by test code ####

# A function to print a simple header
report_header() {
    echo 'Sales summary for the department '$1
}

# A function to print a simple, four column, report
report_line() {
    echo '    Operation: '$1' Costs: '$2' Revenues: '$3' Profits: '$4
}

#### Test code of basic stack machine ####

# Initialize the function stack
pend_init

# This is a LIFO stack - note the reversed order of operations
# Also note that arguments contain whitespace and Bash special characters.

# Three report lines pending
pend_func report_line $( printf '%q\n' 'C' '$4,000.34' '$4,000.00' '(-$0.34)' )
pend_func report_line $( printf '%q\n' 'B' '$2,000.00' '$3,000.00' '$1,000.00' )
pend_func report_line $( printf '%q\n' 'A' '$1,000.00' '$5,000.00' '$4,000.00' )

# Single header line pending
pend_func report_header $( printf '%q\n' '"EasyBucks.con" the Spammer' )

# Mark the stack to test pend_drop
pend_mark

# Put three more report lines on the stack
pend_func report_line $( printf '%q\n' 'G' '$4,000.34' '$4,000.00' '(-$0.34)' )
pend_func report_line $( printf '%q\n' 'F' '$2,000.00' '$3,000.00' '$1,000.00' )
pend_func report_line $( printf '%q\n' 'E' '$1,000.00' '$5,000.00' '$4,000.00' )

# Then decide we don't want them
pend_drop

# Now print the remaining report
echo
pend_release
echo

# The true power of this stack comes when the execution of
# a function adds additional pending operations to the stack.

# Using conditional sequences with full backtracking
# will take you into the world of declarative programming.

:<<-'__func_stack_outputs_'

Sales summary for the department "EasyBucks.con" the Spammer
    Operation: A Costs: $1,000.00 Revenues: $5,000.00 Profits: $4,000.00
    Operation: B Costs: $2,000.00 Revenues: $3,000.00 Profits: $1,000.00
    Operation: C Costs: $4,000.34 Revenues: $4,000.00 Profits: (-$0.34)

__func_stack_outputs_

exit 0