#! /bin/bash
# $Id: split_ip.bash,v 1.5 2004/09/14 22:30:25 mszick Exp $

:<<-'__split_ip_Doc_'

    Copyright (c) Michael S. Zick, 2004
    License: Unrestricted reuse in any form, for any purpose.
    Warranty: None -{Its a script; the user is on their own.}-

Provides
    Given a string representing a properly formed IP(v4) address;
    Split address at the dots;
    Return the normal ordered octets in < array_name_norm >;
    and (optionally)
    Return the reversed ordered octets in < array_name_rev>.

    The reverse ordered return is optional.

__split_ip_Doc_

# Special IFS settings used for string parsing.
# Whitespace == :Space:Tab:Line Feed:Carriage Return:
WSP_IFS=$'\x20'$'\x09'$'\x0A'$'\x0D'

# No Whitespace == Line Feed:Carriage Return
NO_WSP=$'\x0A'$'\x0D'

# Field separator for dotted decimal IP addresses
ADR_IFS=${No_WSP}'.'

####

# Split a properly formed IP address into two arrays.

# The array named as $2 is the normal ordered array;

# The array named as $3 is the reverse ordered array;
# The reverse ordered array is optional.

# split_ip <IP_address> <array_name_norm> [<array_name_rev>]
split_ip() {
    [ $# -eq 3 ] ||              # Either three
    [ $# -eq 2 ] || return 1     # or two arguments

    # Parse by assignment into an array
    local -a _si_input
    local IFS=${ADR_IFS}
    _si_input=( $1 )
    IFS=${WSP_IFS}

    # 'Copy by name' as the normal order array
    eval $2=\(\ \$\{_si_input\[@\]\}\ \)

    if [ $# -eq 3 ]
    then
        # Build query order array
        local -a _dns_ip
        _dns_ip[0]=${_si_input[3]}
        _dns_ip[1]=${_si_input[2]}
        _dns_ip[2]=${_si_input[1]}
        _dns_ip[3]=${_si_input[0]}
        eval $3=\(\ \$\{_dns_ip\[@\]\}\ \)
    fi
    return 0
}

#### Puzzle 1:

# How do you figure out what to escape in an eval statement?
# Such as those 'copy by name' assignments in functions.

# Use the 'quoted' printf format and a hard quoted string:
echo
printf '%q' 'Dst=( ${Src[@]} )'
echo

# The output of that looks like:
# Dst=\(\ \$\{Src\[@\]\}\ \)

# Which gets hand edited into things like:
# eval $3=\(\ \$\{_dns_ip\[@\]\}\ \)

#### Test code:

# Arrays to hold the output
declare -a frwd_octets
declare -a revd_octets

split_ip 10.20.30.40 frwd_octets revd_octets

# Did it work?
echo
echo -n 'Forward array of '${#frwd_octets[@]}' octets '
echo '<'${frwd_octets[@]}'>'

# And the reversed
echo
echo -n 'Reverse array of '${#revd_octets[@]}' octets '
echo '<'${revd_octets[@]}'>'

# How about the two argument version?

# Clear contents of all array elements.
unset frwd_octets[@]
unset revd_octets[@]

split_ip 1.2.3.4 frwd_octets

# Did it work?
echo
echo -n 'Forward array of '${#frwd_octets[@]}' octets '
echo '<'${frwd_octets[@]}'>'

# And the reversed
echo
echo -n 'Reverse array of '${#revd_octets[@]}' octets '
echo '<'${revd_octets[@]}'>'

#### Puzzle 2

# Why are there leading underscores in the variable
# names of a function?
# They are declared to be local, so why?

# The answer is because the functions do 'copy by name'
# evaluations without checking if the passed name is the
# same as an internal name.

# If the caller of the function accidentally provides an
# external name which happens to be the same as an internal
# name, then the function would not behave as expected.

# The following example breaks the above function
# just as a demonstration.

# Accidentally name a global the same as a
# function local...
declare -a _si_input

# Now pass the function the name of the
# global array, and expect things to work:

# set -vx    # In case you want to see how it breaks.

split_ip 4.3.2.1 _si_input
# set +vx

echo
echo -n 'The broken version gives '${#_si_input[@]}
echo    ' octets <'${_si_input[@]}'>'
echo

:<<-'__split_ip_outputs_'

Dst=\(\ \$\{Src\[@\]\}\ \)

Forward array of 4 octets <10 20 30 40>

Reverse array of 4 octets <40 30 20 10>

Forward array of 4 octets <1 2 3 4>

Reverse array of 0 octets <>

The broken version gives 0 octets <>

__split_ip_outputs_

exit 0