#! /bin/bash
# $Id: unique_lines.bash,v 1.1 2004/09/21 14:31:33 mszick Exp $

:<<-'__unique_lines_Doc_'

    Copyright (c) Michael S. Zick, 2004
    License: Unrestricted reuse in any form, for any purpose.
    Warranty: None -{Its a script; the user is on their own.}-

Provides
    A line per array element duplicate remover.

    Array elements may have embedded whitespace
    and special Bash characters.

    Element order is maintained.

    Input and Output array names may be the same;
    allowing an 'edit-in-place' operation.

__unique_lines_Doc_

# Special IFS settings used for string parsing.
# Whitespace == :Space:Tab:Line Feed:Carriage Return:
WSP_IFS=$'\x20'$'\x09'$'\x0A'$'\x0D'

# No Whitespace == Line Feed:Carriage Return
NO_WSP=$'\x0A'$'\x0D'

# unique_lines <in_name> <out_name>
unique_lines() {
    [ $# -eq 2 ] || return 1      # must have 2 arguments

    # Local copy of input array
    local -a _ul_in

    # Local copy of output array
    local -a _ul_out

    # Original element count
    local -i _ul_cnt

    # Current position
    local -i _ul_pos

    # Element content temporary
    local _ul_tmp

    # Line-read mode
    local IFS=${NO_WSP}

    # No glob expansion
    set -f

    # Input array subscript pack by assignment
    eval _ul_in=\( \$\{$1\[@\]\} \)

    # Original element count
    _ul_cnt=${#_ul_in[@]}

    # For each array subscript of input
    for (( _ul_pos = 0 ; _ul_pos < ${_ul_cnt} ; _ul_pos++ ))
    do
        # If element defined and not empty
        if [ ${_ul_in[${_ul_pos}]:=} ]
        then
            # Keep copy for testing
            _ul_tmp=${_ul_in[${_ul_pos}]}

            # Add found element to output array
            _ul_out[${#_ul_out[@]}]=${_ul_tmp}

            # Starting at current position in array
            for (( zap = _ul_pos ; zap < ${_ul_cnt} ; zap++ ))
            do
                # If defined, not empty, and match found
                [ ${_ul_in[${zap}]:=} ] &&
                [ 'x'${_ul_in[${zap}]} == 'x'${_ul_tmp} ] &&
                    unset _ul_in[${zap}]
            done
        fi
    done

    # Copy back output array (which may be empty)
    eval $2=\( \$\{_ul_out\[@\]\} \)

    # Re-enable globbing
    set +f

    # Drop the local IFS setting
    return 0
}


#### Test code ####

declare -a target
declare -a base

base=( '127.0.0.1 localhost' '192.168.0.1 localnet' '0.0.0.0 unknown.domain' )

IFS=${NO_WSP}                                # Line-per-element mode
target=( ${base[@]} ${base[@]} ${base[@]} )  # Triple of each
IFS=${WSP_IFS}

# If correct, should be the contents of base, order maintained.
unique_lines target target

# What did that do for us?
echo
for (( i = 0 ; i < ${#target[@]} ; i++ ))
do
    echo 'The target list element: '$i' contains: >'${target[$i]}'<'
done
echo

:<<-'_unique_lines_outputs_'

The target list element: 0 contains: >127.0.0.1 localhost<
The target list element: 1 contains: >192.168.0.1 localnet<
The target list element: 2 contains: >0.0.0.0 unknown.domain<

_unique_lines_outputs_

exit 0